"""
Renderer
"""
import warnings
try:
    import bpy
except ModuleNotFoundError:  # noqa F821
    warnings.warn(
        'bpy could not be imported, ' +
        'please run your python program in blender')
import numpy as np
import tempfile
import os
import pandas as pd
from navipy.maths.homogeneous_transformations import compose_matrix
import navipy.maths.constants as constants


class BlenderRender():
    """
    BlenderRender is a small class binding python with blender.
    With BlenderRender one can move the bee to a position, and render what
    the bee see at this position.

    The Bee eye is a panoramic camera with equirectangular projection
    The light rays attaining the eyes are filtered with a gaussian.

    """

    def __init__(self):
        """Initialise the Cyberbee
        ..todo check that TemporaryDirectory is writtable and readable
        """
        # Rendering engine needs to be Cycles to support panoramic
        # equirectangular camera
        bpy.context.scene.render.engine = 'CYCLES'
        bpy.context.scene.render.layers["RenderLayer"].use_pass_z = True
        # Look for object camera
        camera_found = False
        for obj in bpy.context.scene.objects:
            if obj.type == 'CAMERA':
                self.camera = obj
                camera_found = True
                break
        assert camera_found, 'The blender file does not contain a camera'
        # The bee eye is panoramic, and with equirectangular projection
        self.camera.data.type = 'PANO'
        self.camera.data.cycles.panorama_type = 'EQUIRECTANGULAR'
        # Filtering props
        bpy.context.scene.cycles.filter_type = 'GAUSSIAN'
        # Call all set function with default values
        self.camera_fov = [[-90, 90], [-180, 180]]
        self.camera_gaussian_width = 1.5
        self.camera_resolution = [360, 180]
        self.cycle_samples = 30
        # switch on nodes
        # Create render link to OutputFile with Image and Z buffer
        bpy.context.scene.use_nodes = True
        scene = bpy.context.scene
        nodes = scene.node_tree.nodes

        render_layers = nodes['Render Layers']
        output_file = nodes.new("CompositorNodeOutputFile")
        output_file.format.file_format = "OPEN_EXR"
        output_file.file_slots.remove(output_file.inputs[0])
        tmp_fileoutput = dict()
        tmp_fileoutput['Image'] = 'Image'
        tmp_fileoutput['Depth'] = 'Depth'
        tmp_fileoutput['Folder'] = tempfile.TemporaryDirectory().name
        tmp_fileoutput['ext'] = '.exr'
        output_file.file_slots.new(tmp_fileoutput['Image'])
        output_file.file_slots.new(tmp_fileoutput['Depth'])
        output_file.base_path = tmp_fileoutput['Folder']
        scene.node_tree.links.new(
            render_layers.outputs['Image'],
            output_file.inputs['Image']
        )
        if bpy.app.version < (2, 79, 0):
            scene.node_tree.links.new(
                render_layers.outputs['Z'],
                output_file.inputs['Depth']
            )
        else:
            scene.node_tree.links.new(
                render_layers.outputs['Depth'],
                output_file.inputs['Depth']
            )
        self.tmp_fileoutput = tmp_fileoutput

    @property
    def cycle_samples(self):
        """get the samples for rendering with cycle

        :returns: the number of samples used for the rendering
        :rtype: int

        """
        return bpy.context.scene.cycles.samples

    @cycle_samples.setter
    def cycle_samples(self, samples=30):
        """change the samples for rendering with cycle

        :param samples: the number of samples to use when rendering images
        :type samples: int

        """
        if not isinstance(samples, int):
            raise TypeError('samples must be an integer')
        bpy.context.scene.cycles.samples = samples

    @property
    def camera_fov(self):
        """get fov of camera

        :returns: the field of view of the camera as \
        [[minimum latitude, maximum latitude],
                            [minimum longitude, maximum longitude]]
                            (in deg)
        :rtype: np.array

        ..todo Change assert to if -> raise TypeError/KeyError

        """
        if not self.camera.data.type == 'PANO':
            raise TypeError('Camera is not panoramic')
        if not self.camera.data.cycles.panorama_type == 'EQUIRECTANGULAR':
            raise TypeError('The panoramic camera is not equirectangular')
        fov = np.zeros((2, 2))
        fov[0, 0] = np.rad2deg(self.camera.data.cycles.latitude_min)
        fov[0, 1] = np.rad2deg(self.camera.data.cycles.latitude_max)
        fov[1, 0] = np.rad2deg(self.camera.data.cycles.longitude_min)
        fov[1, 1] = np.rad2deg(self.camera.data.cycles.longitude_max)
        return fov

    @camera_fov.setter
    def camera_fov(self, resolution):
        """change the field of view of the panoramic camera

        :param resolution: [[minimum latitude, maximum latitude],
                            [minimum longitude, maximum longitude]]
                            (in deg)
        :type latmin: 2x2 float array or list
        """
        if not (isinstance(resolution, tuple) or
                isinstance(resolution, list) or
                isinstance(resolution, np.ndarray)):
            raise TypeError('resolution must be list,  array, or tuple')
        if not self.camera.data.type == 'PANO':
            raise TypeError('Camera is not panoramic')
        if not self.camera.data.cycles.panorama_type == 'EQUIRECTANGULAR':
            raise TypeError('Camera is not equirectangular')
        self.camera.data.cycles.latitude_min = np.deg2rad(
            resolution[0][0])
        self.camera.data.cycles.latitude_max = np.deg2rad(
            resolution[0][1])
        self.camera.data.cycles.longitude_min = np.deg2rad(
            resolution[1][0])
        self.camera.data.cycles.longitude_max = np.deg2rad(
            resolution[1][1])

    @property
    def camera_gaussian_width(self):
        """get width of the gaussian spatial filter

        :returns: the width of the gaussian filter
        :rtype: float

        """
        return bpy.context.scene.cycles.filter_width

    @camera_gaussian_width.setter
    def camera_gaussian_width(self, gauss_w):
        """change width of the gaussian spatial filter

        :param gauss_w: width of the gaussian filter
        :type gauss_w: float


        """
        if not (isinstance(gauss_w, int) or isinstance(gauss_w, float)):
            raise TypeError('gauss window must be integer or float')
        bpy.context.scene.cycles.filter_width = gauss_w

    @property
    def camera_resolution(self):
        """return camera resolution (x,y)

        :returns: the resolution of the camera along (x-axis,y-axis)
        :rtype: (int,int)

        """
        resolution_x = bpy.context.scene.render.resolution_x
        resolution_y = bpy.context.scene.render.resolution_y
        return resolution_x, resolution_y

    @camera_resolution.setter
    def camera_resolution(self, resolution):
        """change the camera resolution (nb of pixels)

        :param resolution_x: number of pixels along the x-axis of the camera
        :type resolution_x: int
        :param resolution_y: number of pixels along the y-axis of the camera
        :type resolution_y: int
        """
        if not (isinstance(resolution, list) or
                isinstance(resolution, np.ndarray)):
            raise TypeError('resolution list or array')
        bpy.context.scene.render.resolution_x = resolution[0]
        bpy.context.scene.render.resolution_y = resolution[1]
        bpy.context.scene.render.resolution_percentage = 100

    @property
    def image(self):
        """return the last rendered image as a numpy array

        :returns: the image (height,width,nchannel)
        :rtype: a double numpy array

        .. note: A temporary file will be written on the harddrive,
                 due to API blender limitation
        """
        # save image as a temporary file, and then loaded
        # sadly the rendered image pixels can not directly be access
        filename = os.path.join(self.tmp_fileoutput['Folder'],
                                self.tmp_fileoutput['Image'] + '0001' +
                                self.tmp_fileoutput['ext'])
        im_width, im_height = self.camera_resolution
        im = bpy.data.images.load(filename)
        pixels = np.array(im.pixels)
        pixels = pixels.reshape([im_height, im_width, -1])
        # The last channel is the alpha channel
        pixels = pixels[..., :(pixels.shape[2] - 1)]
        return pixels

    @property
    def distance(self):
        """return the last rendered distance map as a numpy array

        :returns: the distance map (height, width)
        :rtype: a double numpy array

        .. note: A temporary file will be written on the harddrive,
                 due to API blender limitation

        .. todo: use @property
                     def distance(self)
        """
        # save image as a temporary file, and then loaded
        # sadly the rendered image pixels can not directly be access
        filename = os.path.join(self.tmp_fileoutput['Folder'],
                                self.tmp_fileoutput['Depth'] + '0001' +
                                self.tmp_fileoutput['ext'])
        im_width, im_height = self.camera_resolution
        im = bpy.data.images.load(filename)
        distance = np.array(im.pixels)
        distance = distance.reshape([im_height, im_width, -1])
        # Distance are channel independent
        distance = distance[:, :, 0]
        return distance[..., np.newaxis]

    def update(self, posorient):
        """assign the position and the orientation of the camera.

        :param posorient: is a 1x6 vector containing:
             *in case of euler angeles the index should be
              ['location']['x']
              ['location']['y']
              ['location']['z']
              [convention][alpha_0]
              [convention][alpha_1]
              [convention][alpha_2]
             **where convention can be:
               rxyz, rxzy, ryxz, ryzx, rzyx, rzxy
             *in case of quaternions the index should be
              ['location']['x']
              ['location']['y']
              ['location']['z']
              [convention]['q_0']
              [convention]['q_1']
              [convention]['q_2']
              [convention]['q_3']
             **where convention can be:
               quaternion
             here the angles are euler rotation around the axis
             specified by scene.camera.rotation_mode
        :type posorient: pandas Series with multi-index
        """
        if isinstance(posorient, pd.Series):
            # set roation mode
            conv_found = False
            index = posorient.index
            convention = index.get_level_values(0)[-1]
            for key, _ in constants._AXES2TUPLE.items():
                if convention == key:
                    conv_found = True
                    break
            if convention == 'quaternion':
                conv_found = True
            if not conv_found:
                msg = 'The convention {} used for the orientation'
                msg = msg.format(convention)
                msg += ' is not supported'
                raise ValueError(msg)
        else:
            raise TypeError(
                'posorient must be of type array, list, or pandas Series')
        # Render
        self.camera.matrix_world = compose_matrix(
            angles=posorient.loc[convention],
            translate=posorient.loc['location'],
            axes=convention)
        bpy.ops.render.render()

    def scene(self, posorient):
        """ update position orientation and return a RGBD image

        :param posorient: is a 1x6 vector containing:
             x,y,z, angle_1, angle_2, angle_3,
             here the angles are euler rotation around the axis
             specified by scene.camera.rotation_mode
        :type posorient: 1x6 double array
        :returns: a (height,width, channel) array here the last channel \
is the distance.
        :rtype: a double numpy array
        """
        self.update(posorient)
        toreturn = np.concatenate((self.image,
                                   self.distance), axis=2)
        ninffound = 0
        for chan_i in range(4):
            cim = toreturn[..., chan_i]
            cmax = cim.max()
            if np.isinf(cmax):
                ninffound += np.sum(np.isinf(cim))
                cmax = cim[np.isinf(cim) == 0].max()
                toreturn[np.isinf(cim) == 1] = cmax
        if ninffound > 0:
            warnings.warn('{} Inf found in image'.format(ninffound))
        return toreturn[..., np.newaxis]
